@@ -4,6 +4,7 @@
 #include "qhttpserverrequest_p.h"
 
 #include <QtHttpServer/qhttpserverrequest.h>
+#include <QtNetwork/qhttpheaders.h>
 
 #include <QtCore/qdebug.h>
 #include <QtCore/qloggingcategory.h>
@@ -11,8 +12,9 @@
 #if QT_CONFIG(ssl)
 #include <QtNetwork/qsslsocket.h>
 #endif
-
-Q_LOGGING_CATEGORY(lc, "qt.httpserver.request")
+#if QT_CONFIG(http)
+#include <QtNetwork/private/qhttp2connection_p.h>
+#endif
 
 QT_BEGIN_NAMESPACE
 
@@ -32,17 +34,7 @@ Q_HTTPSERVER_EXPORT QDebug operator<<(QDebug debug, const QHttpServerRequest &re
     QDebugStateSaver saver(debug);
     debug.nospace() << "QHttpServerRequest(";
     debug << "(Url: " << request.url() << ")";
-    debug << "(Headers: ";
-    auto headers = request.headers();
-    bool firstHeader = true;
-    for (auto i = headers.begin(); i != headers.end(); ++i) {
-        if (firstHeader)
-            firstHeader = false;
-        else
-            debug << ", ";
-        debug << "(" << i->first << ": " << i->second << ")";
-    }
-    debug << ")";
+    debug << "(Headers: " << request.headers() << ")";
     debug << "(RemoteHost: " << request.remoteAddress() << ")";
     debug << "(BodySize: " << request.body().size() << ")";
     debug << ')';
@@ -51,6 +43,32 @@ Q_HTTPSERVER_EXPORT QDebug operator<<(QDebug debug, const QHttpServerRequest &re
 
 #endif
 
+namespace {
+
+QHttpServerRequest::Method parseRequestMethod(QByteArrayView str)
+{
+    if (str == "GET")
+        return QHttpServerRequest::Method::Get;
+    else if (str == "PUT")
+        return QHttpServerRequest::Method::Put;
+    else if (str == "DELETE")
+        return QHttpServerRequest::Method::Delete;
+    else if (str == "POST")
+        return QHttpServerRequest::Method::Post;
+    else if (str == "HEAD")
+        return QHttpServerRequest::Method::Head;
+    else if (str == "OPTIONS")
+        return QHttpServerRequest::Method::Options;
+    else if (str == "PATCH")
+        return QHttpServerRequest::Method::Patch;
+    else if (str == "CONNECT")
+        return QHttpServerRequest::Method::Connect;
+    else
+        return QHttpServerRequest::Method::Unknown;
+}
+
+} // anonymous namespace
+
 /*!
     \internal
 */
@@ -88,25 +106,7 @@ bool QHttpServerRequestPrivate::parseRequestLine(QByteArrayView line)
     parser.setMajorVersion(protocol[5] - '0');
     parser.setMinorVersion(protocol[7] - '0');
 
-    if (requestMethod == "GET")
-        method = QHttpServerRequest::Method::Get;
-    else if (requestMethod == "PUT")
-        method = QHttpServerRequest::Method::Put;
-    else if (requestMethod == "DELETE")
-        method = QHttpServerRequest::Method::Delete;
-    else if (requestMethod == "POST")
-        method = QHttpServerRequest::Method::Post;
-    else if (requestMethod == "HEAD")
-        method = QHttpServerRequest::Method::Head;
-    else if (requestMethod == "OPTIONS")
-        method = QHttpServerRequest::Method::Options;
-    else if (requestMethod == "PATCH")
-        method = QHttpServerRequest::Method::Patch;
-    else if (requestMethod == "CONNECT")
-        method = QHttpServerRequest::Method::Connect;
-    else
-        method = QHttpServerRequest::Method::Unknown;
-
+    method = parseRequestMethod(requestMethod);
     url = QUrl::fromEncoded(requestUrl.toByteArray());
     return true;
 }
@@ -351,6 +351,44 @@ bool QHttpServerRequestPrivate::parse(QIODevice *socket)
     return read != -1;
 }
 
+#if QT_CONFIG(http)
+bool QHttpServerRequestPrivate::parse(QHttp2Stream *socket)
+{
+    parser.clear();
+
+    for (const auto &pair : socket->receivedHeaders()) {
+        if (pair.name == ":method") {
+            method = parseRequestMethod(pair.value);
+        } else if (pair.name == ":scheme") {
+            url.setScheme(QLatin1StringView(pair.value));
+        } else if (pair.name == ":authority") {
+            url.setAuthority(QLatin1StringView(pair.value));
+        } else if (pair.name == ":path") {
+            auto path = QUrl::fromEncoded(pair.value);
+            url.setPath(path.path());
+            url.setQuery(path.query());
+        } else {
+            parser.appendHeaderField(pair.name, pair.value);
+        }
+    }
+
+    if (url.scheme().isEmpty())
+        url.setScheme(u"https"_s);
+
+    if (url.host().isEmpty())
+        url.setHost(u"127.0.0.1"_s);
+
+    if (url.port() == -1)
+        url.setPort(port);
+
+    bodyLength = contentLength(); // cache the length
+
+    body = socket->downloadBuffer().readAll();
+
+    return true;
+}
+#endif
+
 /*!
     \internal
 */
@@ -636,13 +674,21 @@ QHttpServerRequest::Method QHttpServerRequest::method() const
 }
 
 /*!
+    \fn const QHttpHeaders &QHttpServerRequest::headers() const &
+    \fn QHttpHeaders QHttpServerRequest::headers() &&
+
     Returns all the request headers.
 */
-QList<QPair<QByteArray, QByteArray>> QHttpServerRequest::headers() const
+const QHttpHeaders &QHttpServerRequest::headers() const &
 {
     return d->parser.headers();
 }
 
+QHttpHeaders QHttpServerRequest::headers() &&
+{
+    return std::move(d->parser).headers();
+}
+
 /*!
     Returns the body of the request.
 */
